library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity sccb_iface is
  
  port (
    pClk   : in std_logic;
    pReset : in std_logic;

    pSIC    : out std_logic;
    pSIDin  : in  std_logic;
    pSIDout : out std_logic;
    pSIDDir : out std_logic;

    pReq  : in  std_logic;
    pBusy : out std_logic;
    
    pAddr    : in  std_logic_vector(7 downto 0);
    pDataIn  : in  std_logic_vector(7 downto 0);
    pDataOut : out std_logic_vector(7 downto 0);
    pWe      : in  std_logic;

    SCCB_ID : in std_logic_vector(7 downto 0);

    debug : out std_logic_vector(7 downto 0)
    );

end sccb_iface;

architecture RTL of sccb_iface is

  type SccbMainStateType is (MAIN_IDLE, DATA_WRITE, DATA_READ);
  signal state : SccbMainStateType;

  signal counter : unsigned(3 downto 0);

  type SccbSubStateType is (SUB_IDLE, START_CONDITION, STOP_CONDITION, OUTPUT_BUFFER, INPUT_BUFFER);
  signal substate : SccbSubStateType;

  signal substate_busy : std_logic;
  signal sub_counter : unsigned(3 downto 0);
  signal buf_counter : unsigned(3 downto 0);
  signal clk_counter : unsigned(3 downto 0);
  
  signal start_condition_req : std_logic;
  signal stop_condition_req  : std_logic;
  signal output_buffer_req   : std_logic;
  signal input_buffer_req    : std_logic;

  signal buf : std_logic_vector(7 downto 0);
  signal sub_buf : std_logic_vector(7 downto 0);

  signal addr_reg : std_logic_vector(7 downto 0);
  signal data_in_reg : std_logic_vector(7 downto 0);

begin  -- RTL

  MAIN_STMT: process (pClk)
  begin  -- process
    if pClk'event and pClk = '1' then
      if pReset = '1' then
        state <= MAIN_IDLE;
        counter <= (others => '0');
        pBusy <= '0';
      else
        case state is
          
          when MAIN_IDLE =>

            if pReq = '1' then
              pBusy <= '1';
              addr_reg <= pAddr;
              if pWe = '1' then
                state <= DATA_WRITE;
                data_in_reg <= pDataIn;
              else
                state <= DATA_READ;
              end if;
            else
              pBusy <= '0';
            end if;
            counter <= (others => '0');

          when DATA_WRITE =>
            case to_integer(counter) is
              when 0 =>                 -- START CODNDITION
                start_condition_req <= '1';
                counter <= counter + 1;
              when 1 =>                 -- output SCCB_ID after START_CONDITION
                start_condition_req <= '0';
                if start_condition_req = '0' and substate_busy = '0' then
                  output_buffer_req <= '1';
                  counter <= counter + 1;
                  buf <= SCCB_ID(7 downto 1) & '0';
                end if;
              when 2 =>                 -- output ADDR after SCCB_ID
                output_buffer_req <= '0';
                if output_buffer_req = '0' and substate_busy = '0' then
                  output_buffer_req <= '1';
                  counter <= counter + 1;
                  buf <= addr_reg;
                end if;
              when 3 =>                 -- output DATA after ADDR
                output_buffer_req <= '0';
                if output_buffer_req = '0' and substate_busy = '0' then
                  output_buffer_req <= '1';
                  counter <= counter + 1;
                  buf <= data_in_reg;
                end if;
              when 4 =>                 -- STOP_CONDITION after DATA
                output_buffer_req <= '0';
                if output_buffer_req = '0' and substate_busy = '0' then
                  stop_condition_req <= '1';
                  counter <= counter + 1;
                end if;
              when 5 =>                 -- return to IDLE after stop_condition
                stop_condition_req <= '0';
                if stop_condition_req = '0' and substate_busy = '0' then
                  state <= MAIN_IDLE;
                  counter <= (others => '0');
                end if;
              when others => null;
            end case;
            
          when DATA_READ =>
            case to_integer(counter) is
              when 0 =>                 -- START_CONDITION
                start_condition_req <= '1';
                counter <= counter + 1;
              when 1 =>                 -- SCCB_ID after START_CONDITION
                start_condition_req <= '0';
                if start_condition_req = '0' and substate_busy = '0' then
                  output_buffer_req <= '1';
                  counter <= counter + 1;
                  buf <= SCCB_ID(7 downto 1) & '0';
                end if;
              when 2 =>                 -- ADDR after SCCB_ID
                output_buffer_req <= '0';
                if output_buffer_req = '0' and substate_busy = '0' then
                  output_buffer_req <= '1';
                  counter <= counter + 1;
                  buf <= addr_reg;
                end if;
              when 3 =>                 -- STOP_CONDITION after ADDR
                output_buffer_req <= '0';
                if output_buffer_req = '0' and substate_busy = '0' then
                  stop_condition_req <= '1';
                  counter <= counter + 1;
                end if;
              when 4 =>                 -- WAIT after STOP_CONDITION
                stop_condition_req <= '0';
                if stop_condition_req = '0' and substate_busy = '0' then
                  counter <= counter + 1;
                end if;
              when 5 =>                 -- START_CONDITION
                start_condition_req <= '1';
                counter <= counter + 1;
              when 6 =>                 -- SCCB_ID after START_CONDITION
                start_condition_req <= '0';
                if start_condition_req = '0' and substate_busy = '0' then
                  output_buffer_req <= '1';
                  counter <= counter + 1;
                  buf <= SCCB_ID(7 downto 1) & '1';  -- read mode
                end if;
              when 7 =>                 -- INPUT DATA after SCCB_ID
                output_buffer_req <= '0';
                if output_buffer_req = '0' and substate_busy = '0' then
                  input_buffer_req <= '1';
                  counter <= counter + 1;
                end if;
              when 8 =>                 -- STOP_CONDITION after INPUT DATA
                input_buffer_req <= '0';
                if input_buffer_req = '0' and substate_busy = '0' then
                  stop_condition_req <= '1';
                  pDataOut <= sub_buf;
                  counter <= counter + 1;
                end if;
              when 9 =>                 -- return to IDLE after STOP_CONDITION
                stop_condition_req <= '0';
                if stop_condition_req = '0' and substate_busy = '0' then
                  state <= MAIN_IDLE;
                  counter <= (others => '0');
                  pBusy <= '0';
                end if;
              when others => null;
            end case;
          when others => state <= MAIN_IDLE;
        end case;
      end if;
    end if;
  end process MAIN_STMT;

  SUB_STMT: process (pClk)
  begin  -- process
    if pClk'event and pClk = '1' then
      if pReset = '1' then
        substate <= SUB_IDLE;
        sub_counter <= (others => '0');
        clk_counter <= (others => '0');
        buf_counter <= (others => '0');
        substate_busy <= '0';
        
        pSIC    <= '1';
        pSIDout <= '1';
        pSIDdir <= '0'; -- output is available
      else
        
        case substate is
          
          when SUB_IDLE =>
            if start_condition_req = '1' then
              substate <= START_CONDITION;
              substate_busy <= '1';
            elsif stop_condition_req = '1' then
              substate <= STOP_CONDITION;
              substate_busy <= '1';
            elsif output_buffer_req = '1' then
              substate <= OUTPUT_BUFFER;
              substate_busy <= '1';
              sub_buf <= buf;
            elsif input_buffer_req = '1' then
              substate <= INPUT_BUFFER;
              substate_busy <= '1';
            else
              substate_busy <= '0';
            end if;
            sub_counter <= (others => '0');
            buf_counter <= (others => '0');
            clk_counter <= (others => '0');
          
          -- before this statement, pSIC should be asserted
          when START_CONDITION =>
            pSIDdir <= '0';             -- output is available
            case to_integer(sub_counter) is
              when 0 =>
                pSIC <= '1';
                pSIDout <= '1';
                if to_integer(clk_counter) = 3 then  -- > 600ns
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 1 =>
                pSIC <= '1';
                pSIDout <= '0';             -- sid will be de-asserted during pSIC is high
                if to_integer(clk_counter) = 3 then  -- > 600ns
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 2 =>
                pSIC <= '0';
                sub_counter <= (others => '0');
                clk_counter <= (others => '0');
                substate_busy <= '0';
                substate <= SUB_IDLE;
              when others => null;
            end case;
            
          -- before this statement, pSIC should be de-asserted
          when STOP_CONDITION =>
            pSIDdir <= '0';             -- output is available
            case to_integer(sub_counter) is
              when 0 =>
                pSIC <= '0';
                pSIDout <= '0';
                if to_integer(clk_counter) = 3 then  -- > 600ns
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 1 =>
                pSIC <= '1';
                pSIDout <= '0';
                if to_integer(clk_counter) = 3 then  -- > 600ns
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 2 =>
                pSIC <= '1';
                pSIDout <= '1';             -- sid will be asserted during pSIC is high
                if to_integer(clk_counter) = 8 then  -- > 1.3us
                  clk_counter <= (others => '0');
                  sub_counter <= (others => '0');
                  substate_busy <= '0';
                  substate <= SUB_IDLE;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when others => null;
            end case;

          -- before this statement, pSIC should be de-asserted
          when OUTPUT_BUFFER =>
            case to_integer(sub_counter) is
              when 0 =>
                pSIDdir <= '0';             -- output is available
                pSIC <= '0';
                if to_integer(clk_counter) = 1 then
                  pSIDout <= sub_buf(7);          -- sid should be changed during pSIC is low
                  sub_buf <= sub_buf(6 downto 0) & '0';
                end if;
                if to_integer(clk_counter) = 8 then  -- > 1.3us
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 1 =>
                pSIDdir <= '0';             -- output is available
                pSIC <= '1';             -- data is available at this state
                if to_integer(clk_counter) = 3 then  -- > 600ns
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 2 =>
                pSIDdir <= '0';             -- output is available
                pSIC <= '0';
                if to_integer(buf_counter) = 7 then
                  buf_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  buf_counter <= buf_counter + 1;
                  sub_counter <= (others => '0');
                end if;
              when 3 =>
                pSIDdir <= '1';             -- input is available
                sub_counter <= sub_counter + 1;
              when 4 =>
                pSIC <= '0';             -- 1 clock to receive 1-bit for ack/nack
--                status <= pSIDin;
                if to_integer(clk_counter) = 8 then  -- > 1.3us
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 5 =>
                pSIC <= '1';
                if to_integer(clk_counter) = 3 then  -- > 600ns
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 6 =>
                pSIC <= '0';
                sub_counter <= sub_counter + 1;
              when 7 =>
                pSIDdir <= '0';             -- output is available
                sub_counter <= (others => '0');
                substate_busy <= '0';
                substate <= SUB_IDLE;
              when others => null;
            end case;
            
          -- before this statement, pSIC should be de-asserted
          when INPUT_BUFFER =>
            case to_integer(sub_counter) is
              when 0 =>
                pSIDdir <= '1'; -- input is available
                pSIC <= '0';
                if to_integer(clk_counter) = 8 then  -- > 1.3us
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 1 =>
                pSIC <= '1';             -- data is available at this state
                sub_counter <= sub_counter + 1;
              when 2 =>
                if to_integer(clk_counter) = 2 then
                  sub_buf <= sub_buf(6 downto 0) & pSIDin;
                end if;
                if to_integer(clk_counter) = 3 then  -- > 600ns
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 3 =>
                pSIC <= '0';
                if to_integer(buf_counter) = 7 then
                  sub_counter <= sub_counter + 1;
                  buf_counter <= (others => '0');
                else
                  sub_counter <= (others => '0');
                  buf_counter <= buf_counter + 1;
                end if;
              when 4 =>
                pSIC <= '0';             -- 1 clock to receive 1-bit for ack/nack
--                status <= pSIDin;
                if to_integer(clk_counter) = 8 then  -- > 1.3us
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 5 =>
                pSIC <= '1';
                if to_integer(clk_counter) = 3 then  -- > 600ns
                  clk_counter <= (others => '0');
                  sub_counter <= sub_counter + 1;
                else
                  clk_counter <= clk_counter + 1;
                end if;
              when 6 =>
                pSIC <= '0';
                sub_counter <= sub_counter + 1;
              when 7 =>
                pSIDdir <= '0';             -- output is available
                sub_counter <= (others => '0');
                substate_busy <= '0';
                substate <= SUB_IDLE;
              when others => null;
            end case;
          when others => substate <= SUB_IDLE;
        end case;
      end if;
    end if;
  end process SUB_STMT;

end RTL;
