library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

library work;
use work.e7cores_pkg.all;

entity top is

  generic (
    MyIpAddr       : std_logic_vector(31 downto 0) := X"0a000001";
    MyMacAddr      : std_logic_vector(47 downto 0) := X"001b1aee001a";
    DefaultGateway : std_logic_vector(31 downto 0) := X"0a0000fe";
    MyNetmask      : std_logic_vector(31 downto 0) := X"ff000000";
    MySendPort0    : std_logic_vector(15 downto 0) := X"4000";
    MySendPort1    : std_logic_vector(15 downto 0) := X"4001";
    TargetIpAddr   : std_logic_vector(31 downto 0) := X"0a000003"
    );

  port (
    pCLK25M          : in  std_logic;
    pReset_n         : in  std_logic;

    -- GMII PHY : 2.5V
    ETH_RESET_N : out   std_logic;  -- このchipにおけるシステムリセットを出す
    -- MAC I/F
    ETH_COL     : inout std_logic;  -- CollisionDetect/MAC_FREQ兼用ピン。外部で2kPUされている。＝＞MAC_CLK出力を125MHzにしている様。
    ETH_CRS     : inout std_logic;  -- CarrierSence/RGMII_SEL0兼用ピン。外部で2kPDされている。
    ETH_TX_CLK  : in    std_logic;  -- TXCLK/RGMII_SEL1兼用ピン。外部で2kPDされている。＝＞CRSと合わせてGMIIモードにしている。
    -- TX out
    ETH_TX_D    : out   std_logic_vector(3 downto 0);  -- 送信データ
    ETH_TX_EN   : out   std_logic;
    -- RX in
    ETH_RX_D    : inout std_logic_vector(3 downto 0);  -- 受信データ
    ETH_RX_CLK  : in    std_logic;  -- 10M=>2.5MHz, 100M=>25MHz, 1G=>125MHz のクロックが来る
    ETH_RX_DV   : inout std_logic;
    ETH_RX_ER   : inout std_logic;
    -- Management I/F
    ETH_MDC     : out   std_logic;  -- クロック出力。max.2.5MHz。
    ETH_MDIO    : inout std_logic;  -- コントロールデータ。
    ETH_PDW     : out   std_logic;

    OV7670_PCLK       : in    std_logic;
    OV7670_PCLK_DUMMY : out   std_logic;
    OV7670_XCLK       : out   std_logic;
    OV7670_VSYNC      : in    std_logic;
    OV7670_HREF       : in    std_logic;
    OV7670_DATA       : in    std_logic_vector(7 downto 0);
    OV7670_SIOC       : out   std_logic;
    OV7670_SIOD       : inout std_logic;

    LED : out std_logic_vector(3 downto 0)
    
    );
end top;

architecture RTL of top is

  component ov7670_capture
    port(
      pPCLK        : in  std_logic;
      pVSYNC       : in  std_logic;
      pHREF        : in  std_logic;
      pDin         : in  std_logic_vector(7 downto 0);
      pAddr        : out std_logic_vector(9 downto 0);
      pDout        : out std_logic_vector(31 downto 0);
      pWe0         : out std_logic;
      pWe1         : out std_logic;
      pLineCounter : OUT std_logic_vector(15 downto 0)
      );
  end component;

  component pll
    port (
      CLK_IN1  : in  std_logic;
      CLK_OUT1 : out std_logic;
      CLK_OUT2 : out std_logic;
      RESET    : in  std_logic;
      LOCKED   : out std_logic
      );
  end component;

  component simpledualportram_dual_clk
    generic (
      DEPTH : integer := 10;
      WIDTH : integer := 32
      );
    port (
      wclk  : in  std_logic;
      rclk  : in  std_logic;
      we    : in  std_logic_vector(0 downto 0);
      raddr : in  std_logic_vector(DEPTH-1 downto 0);
      rdata : out std_logic_vector(WIDTH-1 downto 0);
      waddr : in  std_logic_vector(DEPTH-1 downto 0);
      wdata : in  std_logic_vector(WIDTH-1 downto 0)
      );
  end component;

  component sccb_iface
    port (
      pClk   : in std_logic;
      pReset : in std_logic;

      pSIC    : out std_logic;
      pSIDin  : in  std_logic;
      pSIDout : out std_logic;
      pSIDdir : out std_logic;

      pReq  : in  std_logic;
      pBusy : out std_logic;
      pAddr : in std_logic_vector(7 downto 0);
      
      pDataIn  : in  std_logic_vector(7 downto 0);
      pDataOut : out std_logic_vector(7 downto 0);
      pWe      : in  std_logic;

      SCCB_ID : in std_logic_vector(7 downto 0);

      debug : out std_logic_vector(7 downto 0)
      );
  end component;

  component udp_sccb_bridge
    port (
      pClk   : in std_logic;
      pReset : in std_logic;
      
      pO0Data    : out std_logic_vector(31 downto 0);
      pO0Request : out std_logic;
      pO0Ack     : in  std_logic;
      pO0Enable  : out std_logic;

      pI0Data    : in std_logic_vector(31 downto 0);
      pI0Request : in std_logic;
      pI0Ack     : out  std_logic;
      pI0Enable  : in std_logic;

      pSccbAddr    : out std_logic_vector(7 downto 0);
      pSccbDataIn  : out std_logic_vector(7 downto 0);
      pSccbDataOut : in  std_logic_vector(7 downto 0);
      pSccbDataWe  : out std_logic;
      
      pSccbReq  : out std_logic;
      pSccbBusy : in  std_logic;
      
      debug : out std_logic_vector(7 downto 0)
      );
    
  end component;

  component ov7670_udp_bridge
    port (
      reset : in std_logic;
      
      pOV7670_PixelClk : in std_logic;
      pLineCounter     : in std_logic_vector(15 downto 0);
      
      pClk_50M  : in  std_logic;
      pMemAddr  : out std_logic_vector(9 downto 0);
      pMemData0 : in  std_logic_vector(31 downto 0);
      pMemData1 : in  std_logic_vector(31 downto 0);
      
      pO0Data    : out std_logic_vector(31 downto 0);
      pO0Request : out std_logic;
      pO0Ack     : in  std_logic;
      pO0Enable  : out std_logic;
    
      pSrcIPAddr  : in std_logic_vector(31 downto 0);
      pDestIPAddr : in std_logic_vector(31 downto 0);
      pSrcPort    : in std_logic_vector(15 downto 0);
      pDestPort   : in std_logic_vector(15 downto 0);
    
      debug : out std_logic_vector(7 downto 0)
      );
  end component;

  signal clk_50M  : std_logic;
  signal clk_25M  : std_logic;
  
  signal we0, we1 : std_logic;
  
  signal waddr : std_logic_vector(9 downto 0);
  signal wdata : std_logic_vector(31 downto 0);
  
  signal raddr  : std_logic_vector(9 downto 0);
  signal rdata0 : std_logic_vector(31 downto 0);
  signal rdata1 : std_logic_vector(31 downto 0);
  
  type UPL32 is record
    Data    : std_logic_vector(31 downto 0);
    Enable  : std_logic;
    Request : std_logic;
    Ack     : std_logic;
  end record;

  signal Udp0Send    : UPL32;
  signal Udp1Send    : UPL32;
  signal Udp1Receive : UPL32;
  
  signal sEMAC25MHz : std_logic;
  
  signal line_counter      : std_logic_vector(15 downto 0);

  signal sccb_addr     : std_logic_vector(7 downto 0);
  signal sccb_data_in  : std_logic_vector(7 downto 0);
  signal sccb_data_out : std_logic_vector(7 downto 0);
  signal sccb_data_we  : std_logic;
  signal sccb_req      : std_logic;
  signal sccb_busy     : std_logic;

  signal SDA_I   : std_logic;
  signal SDA_O   : std_logic;
  signal SDA_DIR : std_logic;
  signal SCL     : std_logic;

  signal local_clock_counter : std_logic_vector(15 downto 0);
  signal xclk : std_logic;

  signal reset : std_logic;

begin

  reset <= not pReset_n;
  
  ETH_PDW <= '1';
  
  OV7670_PCLK_DUMMY <= 'Z';

  LED(0) <= '0';
  LED(1) <= '1';
  LED(2) <= '1';
  LED(3) <= '0';
  
  U_PLL : pll port map (
    CLK_IN1  => pCLK25M,
    CLK_OUT1 => clk_50M,
    CLK_OUT2 => clk_25M,
    LOCKED   => open,
    RESET    => '0'
    );

  ov7670_xclk <= xclk;
  process(clk_50M)
  begin
    if clk_50M'event and clk_50M = '1' then
      xclk <= not xclk;
    end if;
  end process;

  U_BUF0 : simpledualportram_dual_clk
    generic map(
      DEPTH => 10,
      WIDTH => 32
      )
    port map(
      wclk  => OV7670_PCLK,
      rclk  => clk_50M,
      we(0) => we0,
      raddr => raddr,
      rdata => rdata0,
      waddr => waddr,
      wdata => wdata
      );

  U_BUF1 : simpledualportram_dual_clk
    generic map(
      DEPTH => 10,
      WIDTH => 32
      )
    port map(
      wclk  => OV7670_PCLK,
      rclk  => clk_50M,
      we(0) => we1,
      raddr => raddr,
      rdata => rdata1,
      waddr => waddr,
      wdata => wdata
      );

  U_OV7670_CAPTURE : ov7670_capture
    port map(
      pPCLK        => OV7670_PCLK,
      pVSYNC       => OV7670_VSYNC,
      pHREF        => OV7670_HREF,
      pDin         => OV7670_DATA,
      pAddr        => waddr,
      pDout        => wdata,
      pWe0         => we0,
      pWe1         => we1,
      pLineCounter => line_counter
      );

  U_OV7670_UDP_BRIDGE: ov7670_udp_bridge
    port map(
      reset => reset,
      
      pOV7670_PixelClk  => OV7670_PCLK,
      pLineCounter => line_counter,
      
      pClk_50M  => clk_50M,
      pMemAddr  => raddr,
      pMemData0 => rdata0,
      pMemData1 => rdata1,
      
      pO0Data    => Udp0Send.Data,
      pO0Request => Udp0Send.Request,
      pO0Ack     => Udp0Send.Ack,
      pO0Enable  => Udp0Send.Enable,

      pSrcIPAddr  => MyIpAddr,
      pDestIPAddr => TargetIpAddr,
      pSrcPort    => MySendPort0,
      pDestPort   => MySendPort0,
      
      debug => open
      );

  sEMAC25MHz <= CLK_25M;
  u_e7udpip0 : e7udpip100M_exstick
    port map (
      -- GMII PHY : 2.5V
      EPHY_RST_N    => ETH_RESET_N,
      EPHY_MAC_CLK  => sEMAC25MHz, -- USER_CLOCK,
      -- MAC I/F
      EPHY_COL      => ETH_COL,
      EPHY_CRS      => ETH_CRS,
      EPHY_TXCLK    => ETH_TX_CLK,
      -- TX out
      EPHY_TD       => ETH_TX_D,
      EPHY_TXEN     => ETH_TX_EN,
      EPHY_TXER     => open,
      -- RX in
      EPHY_RD       => ETH_RX_D,
      EPHY_RXCLK    => ETH_RX_CLK,
      EPHY_RXDV     => ETH_RX_DV,
      EPHY_RXER     => ETH_RX_ER,
      -- Management I/F
      EPHY_MDC      => ETH_MDC,
      EPHY_MDIO     => ETH_MDIO,
      EPHY_INT_N    => '1',

      -- Asynchronous Reset
      Reset_n       => pReset_n,
      -- UPL interface
      pUPLGlobalClk => clk_50M,
      -- UDP tx input
      pUdp0Send_Data    => Udp0Send.Data,
      pUdp0Send_Request => Udp0Send.Request,
      pUdp0Send_Ack     => Udp0Send.Ack,
      pUdp0Send_Enable  => Udp0Send.Enable,

      pUdp1Send_Data    => Udp1Send.Data,
      pUdp1Send_Request => Udp1Send.Request,
      pUdp1Send_Ack     => Udp1Send.Ack,
      pUdp1Send_Enable  => Udp1Send.Enable,

      -- UDP rx output
      pUdp0Receive_Data    => open,     -- Udp0Receive.Data,
      pUdp0Receive_Request => open,     -- Udp0Receive.Request,
      pUdp0Receive_Ack     => '0',      -- Udp0Receive.Ack,
      pUdp0Receive_Enable  => open,     -- Udp0Receive.Enable,

      pUdp1Receive_Data    => Udp1Receive.Data,
      pUdp1Receive_Request => Udp1Receive.Request,
      pUdp1Receive_Ack     => Udp1Receive.Ack,
      pUdp1Receive_Enable  => Udp1Receive.Enable,

      -- Setup
      pMyIpAddr       => MyIpAddr,
      pMyMacAddr      => MyMacAddr,
      pMyNetmask      => MyNetmask,
      pDefaultGateway => DefaultGateway,
      pMyUdpPort0     => MySendPort0,
      pMyUdpPort1     => MySendPort1,
      pTargetIpAddr   => TargetIpAddr,

      -- Status
      pStatus_RxByteCount             => open,
      pStatus_RxPacketCount           => open,
      pStatus_RxErrorPacketCount      => open,
      pStatus_RxDropPacketCount       => open,
      pStatus_RxARPRequestPacketCount => open,
      pStatus_RxARPReplyPacketCount   => open,
      pStatus_RxICMPPacketCount       => open,
      pStatus_RxUDP0PacketCount       => open,
      pStatus_RxUDP1PacketCount       => open,
      pStatus_RxIPErrorPacketCount    => open,
      pStatus_RxUDPErrorPacketCount   => open,

      pStatus_TxByteCount             => open,
      pStatus_TxPacketCount           => open,
      pStatus_TxARPRequestPacketCount => open,
      pStatus_TxARPReplyPacketCount   => open,
      pStatus_TxICMPReplyPacketCount  => open,
      pStatus_TxUDP0PacketCount       => open,
      pStatus_TxUDP1PacketCount       => open,
      pStatus_TxMulticastPacketCount  => open,
		
		
		pStatus_phy_link_100M       => open,
      pStatus_phy_link_fullduplex => open,
      pStatus_phy_link_up         => open,

      pdebug => open

    );

  IOBUF_inst : IOBUF
     generic map (
       DRIVE      => 12,
       IOSTANDARD => "DEFAULT",
       SLEW       => "SLOW")
     port map (
       O  => SDA_I,       -- Buffer output
       IO => OV7670_SIOD, -- Buffer inout port (connect directly to top-level port)
       I  => SDA_O,  -- Buffer input
       T  => SDA_DIR -- 3-state enable input, high=input, low=output 
       );
  OV7670_SIOC <= SCL;
  
  process (clk_50M)
  begin
    if clk_50M'event and clk_50M = '1' then
      local_clock_counter <= std_logic_vector(unsigned(local_clock_counter) + 1);
    end if;
  end process;

  U_SCCB_IFACE : sccb_iface port map(
    pClk => local_clock_counter(3),
    pReset  => reset,

    pSIC    => SCL,
    pSIDin  => SDA_I,
    pSIDout => SDA_O,
    pSIDdir => SDA_DIR,

    pReq  => sccb_req,
    pBusy => sccb_busy,
    
    pAddr    => sccb_addr,
    pDataIn  => sccb_data_in,
    pDataOut => sccb_data_out,
    pWe      => sccb_data_we,

    SCCB_ID => X"42",

    debug => open
    );
  
  U_UDP_SCCB_BRIDGE : udp_sccb_bridge port map(
    pClk   => clk_50M,
    pReset => reset,
      
    pO0Data    => Udp1Send.Data,
    pO0Request => Udp1Send.Request,
    pO0Ack     => Udp1Send.Ack,
    pO0Enable  => Udp1Send.Enable,

    pI0Data    => Udp1Receive.Data,
    pI0Request => Udp1Receive.Request,
    pI0Ack     => Udp1Receive.Ack,
    pI0Enable  => Udp1Receive.Enable,

    pSccbAddr    => sccb_addr,
    pSccbDataIn  => sccb_data_in,
    pSccbDataOut => sccb_data_out,
    pSccbDataWe  => sccb_data_we,
      
    pSccbReq  => sccb_req,
    pSccbBusy => sccb_busy,
      
    debug => open
    );

end RTL;

