library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;

entity udp_sccb_bridge is

  port (
    pClk   : in std_logic;
    pReset : in std_logic;
    
    pO0Data    : out std_logic_vector(31 downto 0);
    pO0Request : out std_logic;
    pO0Ack     : in  std_logic;
    pO0Enable  : out std_logic;

    pI0Data    : in std_logic_vector(31 downto 0);
    pI0Request : in std_logic;
    pI0Ack     : out  std_logic;
    pI0Enable  : in std_logic;

    pSccbAddr     : out std_logic_vector(7 downto 0);
    pSccbDataIn  : out std_logic_vector(7 downto 0);
    pSccbDataOut : in  std_logic_vector(7 downto 0);
    pSccbDataWe  : out std_logic;
    
    pSccbReq  : out std_logic;
    pSccbBusy : in  std_logic;
    
    debug : out std_logic_vector(7 downto 0)
    );
  
end udp_sccb_bridge;

architecture RTL of udp_sccb_bridge is

  type StateType is (IDLE, RECV, ACCESS_TO_SCCB, WAIT_SCCB_START, WAIT_SCCB_DONE, WAIT_SEND_ACK, SEND_DATA);
  signal state : StateType;

  signal cSrcIPAddr  : std_logic_vector(31 downto 0);
  signal cDestIPAddr : std_logic_vector(31 downto 0);
  signal cSrcPort    : std_logic_vector(15 downto 0);
  signal cDestPort   : std_logic_vector(15 downto 0);

  signal sccb_access_mode : std_logic;
  signal sccb_access_addr : std_logic_vector(7 downto 0);
  signal sccb_access_data : std_logic_vector(7 downto 0);
  signal sccb_req         : std_logic;

  signal rx_counter : std_logic_vector(11 downto 0);
  signal tx_counter : std_logic_vector(11 downto 0);
  

begin  -- RTL

  pSccbReq <= sccb_req;
  
  process (pClk)
  begin  -- process
    if pClk'event and pClk = '1' then
      if pReset = '1' then
        state  <= IDLE;
        pI0Ack <= '0';
        tx_counter <= (others => '0');
        rx_counter <= (others => '0');
      else
        case state is
          
          when IDLE =>
            pO0Enable <= '0';
            if pI0Enable = '1' then
              state  <= RECV;
              cSrcIPAddr <= pI0Data;
              rx_counter <= std_logic_vector(unsigned(rx_counter) + 1);
              pI0Ack <= '0';
            else
              rx_counter <= (others => '0');
              pI0Ack <= '1';
            end if;
            tx_counter <= (others => '0');
            
          when RECV =>
            pI0Ack <= '0';
            if pI0Enable = '0' then
              state <= ACCESS_TO_SCCB;
              rx_counter <= (others => '0');
            else
              rx_counter <= std_logic_vector(unsigned(rx_counter) + 1);
              case to_integer(unsigned(rx_counter)) is
                when 0 => null;         -- 先のステートで消費している
                when 1 => cDestIPAddr <= pI0Data;
                when 2 =>
                  cDestPort <= pI0Data(31 downto 16);
                  cSrcPort  <= pI0Data(15 downto  0);
                when 3 => null;
                when 4 =>
                  sccb_access_mode <= pI0Data(24);
                  sccb_access_addr <= pI0Data(15 downto 8);
                  sccb_access_data <= pI0Data( 7 downto 0);
                  state <= ACCESS_TO_SCCB;
                when others =>
                  state <= ACCESS_TO_SCCB;
              end case;
            end if;
            
          when ACCESS_TO_SCCB =>
            if pSccbBusy = '0' then
              sccb_req    <= '1';
              pSccbDataWe <= sccb_access_mode;
              pSccbAddr   <= sccb_access_addr;
              if sccb_access_mode = '1' then
                pSccbDataIn <= sccb_access_data;
              end if;
              state <= WAIT_SCCB_START;
            end if;

          when WAIT_SCCB_START =>
            if pSccbBusy = '1' then
              sccb_req <= '0';
              state    <= WAIT_SCCB_DONE;
            end if;
            
          when WAIT_SCCB_DONE =>
            if pSccbBusy = '0' then
              state      <= WAIT_SEND_ACK;
              pO0Request <= '1';
            else
              sccb_req <= '0';
            end if;
            tx_counter <= (others => '0');
              
          when WAIT_SEND_ACK =>
            if pO0Ack = '1' then         -- データを送れるようになった
              pO0Request <= '0';
              pO0Enable  <= '1';         -- データの送信開始
              pO0Data    <= cSrcIPAddr;  -- 送り元(自分)のIPアドレス
              tx_counter <= std_logic_vector(unsigned(tx_counter) + 1);
              state      <= SEND_DATA;
            end if;
            
          when SEND_DATA =>
            tx_counter <= std_logic_vector(unsigned(tx_counter) + 1);
            case to_integer(unsigned(tx_counter)) is
              when 0 => null;                              -- tx_counterが1に相当する送り元のIPアドレスは先のステートで送信済み
              when 1 => pO0Data <= cDestIPAddr;            -- 送り先のIPアドレス
              when 2 => pO0Data <= cDestPort & cSrcPort;   -- 送り元のポート番号
              when 3 =>                 -- length
                pO0Data <= X"00000004";
              when 4 =>                 -- data
                pO0Data(31 downto 24) <= "0000000" & "1";
                pO0Data(23 downto 8) <= (others => '0');
                if sccb_access_mode = '0' then
                  pO0Data(7 downto 0) <= pSccbDataOut;
                else
                  pO0Data(7 downto 0) <= (others => '0');
                end if;
                state <= IDLE;
              when others => null;
            end case;
            
          when others =>
            state <= IDLE;
        end case;
        
      end if;
    end if;
  end process;
  
end RTL;
